using System;
using System.Linq;
using System.Collections;
using System.Collections.Generic;
using System.Text.RegularExpressions;
using MeshVR;
using MVR.FileManagement;
using MVR.FileManagementSecure;

using UnityEngine;
using UnityEngine.UI;
using UnityEngine.Rendering;
using UnityEngine.PostProcessing;

namespace moyashi {  
  public class PostProcessing : MVRScript {
    private static uint VERSION = 7;

    private EffectsCamera effectsCamera;

    // builtin cameras
    private Camera screenshotCamera;
    private Camera thumbnailCamera;

    private UIDynamicButton customLutButton;
    private UIDynamicTextField customLutTextfield;

    private ImageLoaderThreaded.QueuedImage queuedImage;
    private Texture2D texture;

    List<UIDynamicButton> tabs = new List<UIDynamicButton>();
    List<UIDynamic> colorGradingTab = new List<UIDynamic>();
    List<UIDynamic> lensTab = new List<UIDynamic>();
    List<UIDynamic> atmosphereTab = new List<UIDynamic>();

    private string getPackagePath() {
      string filename = manager.GetJSON()["plugins"][name.Split('_')[0]].Value;
      int index = filename.IndexOf(":/");

      if (index > 0) {
        string path = filename.Substring(0, index + 2);
        // SuperController.LogMessage($"package path: {path}");
        return path;
      } else {
        return "";
      }
    }

    // fuck this is horrible surely theres a better way
    private class AssetLoadedMonitor : CustomYieldInstruction {
      private bool finished = false;

      public override bool keepWaiting { 
        get { return !finished; }
      }

      public void finish(AssetLoader.AssetBundleFromFileRequest fileRequest) {
        // SuperController.LogMessage("finish");
        this.finished = true;
        
        if (fileRequest.assetBundle == null) {
          AssetLoader.DoneWithAssetBundleFromFile(fileRequest.path);
        }
      }
    }

    private IEnumerator loadShaders() {
      AssetLoadedMonitor assetLoadedMonitor = new AssetLoadedMonitor();
      AssetLoader.AssetBundleFromFileRequest fileRequest = new AssetLoader.AssetBundleFromFileRequest {
        path = getPackagePath() + $"Custom/Assets/moyashi/Moyashi_PostProcessing{VERSION}.assetbundle",
        callback = assetLoadedMonitor.finish
      };

      // TODO: investigate why we arent calling done enough...
      AssetLoader.DoneWithAssetBundleFromFile(fileRequest.path);
      AssetLoader.DoneWithAssetBundleFromFile(fileRequest.path);
      AssetLoader.DoneWithAssetBundleFromFile(fileRequest.path);
      AssetLoader.DoneWithAssetBundleFromFile(fileRequest.path);
      AssetLoader.QueueLoadAssetBundleFromFile(fileRequest);
      yield return StartCoroutine(assetLoadedMonitor);

      if (fileRequest.assetBundle != null) {
        Shader prepassShader = fileRequest.assetBundle.LoadAsset<Shader>($"MoyashiPrepass{VERSION}.shader");
        Shader colorCorrectionShader = fileRequest.assetBundle.LoadAsset<Shader>($"MoyashiColorCorrection{VERSION}.shader");
        Shader passthroughShader = fileRequest.assetBundle.LoadAsset<Shader>($"Passthrough{VERSION}.shader");
        ComputeShader computeShader = fileRequest.assetBundle.LoadAsset<ComputeShader>($"MoyashiCompute{VERSION}.compute");
        ComputeShader lensComputeShader = fileRequest.assetBundle.LoadAsset<ComputeShader>($"MoyashiLens{VERSION}.compute");

        if (effectsCamera != null) {
          // SuperController.LogMessage($"adding shader to {customPostProcessingSet.Count} custom post processors");

          // TODO: solution for other cameras
          // foreach(CustomPostProcessing cpp in customPostProcessingSet) {
            // cpp.setColorCorrectionShader(colorCorrectionShader);
            // cpp.setPassthroughShader(passthroughShader);
            // cpp.setComputeShader(computeShader);
            // cpp.setLensComputeShader(lensComputeShader);
          // }

          // TODO: get rid of this
          effectsCamera.setPrepassShader(prepassShader);
          effectsCamera.setShaders(colorCorrectionShader, computeShader, lensComputeShader);
        } else {
          SuperController.LogError("effects camera was null when trying to load shaders, this shouldnt be possible");
        }
      } else {
        SuperController.LogError("assetbundle failed to load");
      }
    }

    public void Defer(Action f, float delay) {
      StartCoroutine(InvokeRoutine(f, delay));
    }
    
    private IEnumerator InvokeRoutine(System.Action f, float delay) {
      yield return new WaitForSeconds(delay);
      f();
    }

    // applies a setting to all instances of CustomPostProcessing... might change how
    // this works but it's pretty convenient
    private void applySetting(Action<EffectsCamera> f) {
      // TODO: rework
      // foreach (CustomPostProcessing customPostProcessing in customPostProcessingSet) {
      //   f(customPostProcessing);
      // }

      f(effectsCamera);
    }

    private UIDynamicSlider CreateMiniDynamicSlider(string label, JSONStorableFloat f) {
      UIDynamicSlider slider = CreateSlider(f, false);   
      slider.Configure(label, f.min, f.max, f.val, true, "F3", false, false);   
      slider.defaultButtonEnabled = false;
      slider.quickButtonsGroup.gameObject.SetActive(false);
      // slider.sliderValueTextFromFloat.gameObject.SetActive(false);
      LayoutElement component = slider.GetComponent<LayoutElement>();
      if (component != null) {
        component.minHeight = 0.0f;
        component.preferredHeight = 0.0f;
        component.flexibleHeight = 0.0f;
        component.ignoreLayout = true; // ideally ignore layout, but causes it to spring back to normal height? annoying
        // component.minHeight = 50.0f;
        // component.preferredHeight = 50.0f;
      }

      RectTransform rect = slider.GetComponent<RectTransform>();
      Vector2 offset = rect.offsetMin;
      offset.x = 0.0f;
      offset.y = 0.0f;
      rect.offsetMin = offset;
      rect.sizeDelta = new Vector2(0.0f, 0.0f);

      // turn slider into "mini bar"
      RectTransform rect2 = slider.slider.GetComponent<RectTransform>();
      if (rect2 != null) {
        Vector2 sizeDelta = rect2.sizeDelta;
        // sizeDelta.y = 40.0f;
        sizeDelta.y = 0.0f;
        rect2.sizeDelta = sizeDelta;
        rect2.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 0, 430.0f);
        rect2.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 5.0f, 40.0f);

        Vector2 offsetMin = rect2.offsetMin;
        offsetMin.x = 40.0f;
        rect2.offsetMin = offsetMin;
      }

      // make sure the clickable text input doesnt overlap the mini slider bar
      slider.sliderValueTextFromFloat.UIInputField.textComponent.rectTransform.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Right, 0, 85.0f);

      return slider;
    }

    // TODO: probably move this to its own class so we can destruct the custom text elements etc
    private UIDynamic createVectorMixer(JSONStorableVector3 v, bool alignRight) {
      UIDynamic spacer = CreateSpacer(alignRight);
      Transform mixer = spacer.transform;
      // RectTransform rect = mixer.GetComponent<RectTransform>();
      // if (rect != null) {
      //   rect.
      // }
      spacer.height = 178.0f;

      UIContainer container = mixer.gameObject.AddComponent<UIContainer>();

      JSONStorableString storable = new JSONStorableString("_Label", v.name);
      UIDynamicTextField textfield = CreateTextField(storable, alignRight);
      textfield.textColor = new Color(1,1,1);
      textfield.height = 30.0f;
      textfield.UItext.transform.parent = mixer;
      textfield.UItext.fontSize = 24;
      textfield.UItext.fontStyle = FontStyle.Bold;
      textfield.UItext.alignment = TextAnchor.MiddleCenter;

      LayoutElement component = textfield.GetComponent<LayoutElement>();
      component.ignoreLayout = true;

      RectTransform textfieldRect = textfield.GetComponent<RectTransform>();
      textfieldRect.sizeDelta = new Vector2(0.0f, 0.0f);

      RectTransform textRect = textfield.UItext.GetComponent<RectTransform>();
      textRect.localPosition = new Vector3(0,0,0);
      textRect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 2.5f, 30.0f);
      textRect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 10.0f, 500.0f);

      // default button
      UIDynamicButton defaultButton = CreateButton("Default", alignRight);
      defaultButton.buttonText.fontSize = 16;

      // dunno why you need to defer all this layout stuff but whatever
      Defer(() => {
        LayoutElement defaultButtonLayout = defaultButton.GetComponent<LayoutElement>();
        defaultButtonLayout.ignoreLayout = true;
        RectTransform defaultButtonRect = defaultButton.GetComponent<RectTransform>();
        defaultButtonRect.parent = container.transform;
        defaultButtonRect.localPosition = new Vector3(0,0,0);
        defaultButtonRect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 2.5f, 30.0f);
        defaultButtonRect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Right, 2.5f, 100.0f);  
      }, 0.1f);

      // attach and re-align sliders
      JSONStorableFloat r = new JSONStorableFloat(v.name + "Red", v.val[0], v.min[0], v.max[0], true, true);
      RegisterFloat(r);
      { // red        
        UIDynamicSlider sliderR = CreateMiniDynamicSlider("R", r);

        Defer(() => {
          sliderR.transform.parent = mixer;
          LayoutElement layout = sliderR.GetComponent<LayoutElement>();
          layout.ignoreLayout = true;
          RectTransform rect = sliderR.GetComponent<RectTransform>();
          rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 0.0f, 525.0f); 
          rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 30.0f, 0.0f);
        }, 0.1f);

        r.setCallbackFunction = (float c) => { v.SetXVal(c); };
      }

      JSONStorableFloat g = new JSONStorableFloat(v.name + "Green", v.val[1], v.min[1], v.max[1], true, true);
      RegisterFloat(g);
      { // green        
        UIDynamicSlider sliderG = CreateMiniDynamicSlider("G", g);

        Defer(() => {
          sliderG.transform.parent = mixer;
          LayoutElement layout = sliderG.GetComponent<LayoutElement>();
          layout.ignoreLayout = true;
          RectTransform rect = sliderG.GetComponent<RectTransform>();
          rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 0.0f, 525.0f); 
          rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 77.5f, 0.0f);
        }, 0.1f);

        g.setCallbackFunction = (float c) => { v.SetYVal(c); };
      }

      JSONStorableFloat b = new JSONStorableFloat(v.name + "Blue", v.val[2], v.min[2], v.max[2], true, true);
      RegisterFloat(b);
      { // blue        
        UIDynamicSlider sliderB = CreateMiniDynamicSlider("B", b);

        Defer(() => {
          sliderB.transform.parent = mixer;
          LayoutElement layout = sliderB.GetComponent<LayoutElement>();
          layout.ignoreLayout = true;
          RectTransform rect = sliderB.GetComponent<RectTransform>();
          rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 0.0f, 525.0f); 
          rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 125.0f, 0.0f);
        }, 0.1f);

        b.setCallbackFunction = (float c) => { v.SetZVal(c); };
      }

      defaultButton.button.onClick.AddListener(() => {
        r.SetValToDefault();
        g.SetValToDefault();
        b.SetValToDefault();
      });

      return spacer;
    }

    private void initChannelMixer(List<UIDynamic> tab) {
      Color initialR = new Color(1.0f, 0.0f, 0.0f);
      Color initialG = new Color(0.0f, 1.0f, 0.0f);
      Color initialB = new Color(0.0f, 0.0f, 1.0f);

      float min = -1.0f;
      float max = 2.0f;

      Vector3 mixerMin = new Vector3(min, min, min);
      Vector3 mixerMax = new Vector3(max, max, max);

      JSONStorableVector3 mixerR = new JSONStorableVector3("Channel mixer (R)", new Vector3(1.0f, 0.0f, 0.0f), mixerMin, mixerMax, true, true);
      JSONStorableVector3 mixerG = new JSONStorableVector3("Channel mixer (G)", new Vector3(0.0f, 1.0f, 0.0f), mixerMin, mixerMax, true, true);
      JSONStorableVector3 mixerB = new JSONStorableVector3("Channel mixer (B)", new Vector3(0.0f, 0.0f, 1.0f), mixerMin, mixerMax, true, true);

      UIDynamic vectorMixerR = createVectorMixer(mixerR, false);
      UIDynamic vectorMixerG = createVectorMixer(mixerG, false);
      UIDynamic vectorMixerB = createVectorMixer(mixerB, false);

      mixerR.setCallbackFunction = (Vector3 v) => { applySetting((s) => s.setChannelMixR(v)); };
      mixerG.setCallbackFunction = (Vector3 v) => { applySetting((s) => s.setChannelMixG(v)); };
      mixerB.setCallbackFunction = (Vector3 v) => { applySetting((s) => s.setChannelMixB(v)); };

      tab.Add(vectorMixerR);
      tab.Add(vectorMixerG);
      tab.Add(vectorMixerB);
    }

    private void initContrastScale(List<UIDynamic> tab) {
      float initialValue = 0.0f;
      float min = -1.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Contrast", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, false);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setContrastScale(v)); };

      tab.Add(slider);
    }

    private void initTonemappingSelect(List<UIDynamic> tab) {
      float initialValue = 1.0f;
      float min = 0.0f;
      float max = 1.0f;

      List<string> tonemappingMethods = new List<string>();
      tonemappingMethods.Add("None");
      tonemappingMethods.Add("ACES");

      JSONStorableStringChooser storableTonemap = new JSONStorableStringChooser("Tonemapper", tonemappingMethods, "ACES", "Tonemap");
      UIDynamicPopup dropdown = CreatePopup(storableTonemap, true);

      JSONStorableFloat storableTonemapStrength = new JSONStorableFloat("Tonemapping strength", initialValue, min, max, true, true);
      storableTonemapStrength.storeType = JSONStorableParam.StoreType.Full;

      storableTonemap.setCallbackFunction = (string s) => {
        switch(s) {
          case "None":
            storableTonemapStrength.val = 0.0f;
            storableTonemapStrength.max = 0.0f;
            break;
          case "ACES":
            storableTonemapStrength.max = max;
            storableTonemapStrength.val = 1.0f;
            break;
        }
      };

      UIDynamicSlider slider = CreateSlider(storableTonemapStrength, true);      
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storableTonemapStrength);
      storableTonemapStrength.setCallbackFunction = (float v) => { applySetting((s) => s.setTonemappingScale(v)); };

      tab.Add(dropdown);
      tab.Add(slider);
    }

    private void initTonemappingExposure(List<UIDynamic> tab) {
      float initialValue = 1.0f;
      float min = 0.0f;
      float max = 5.0f;

      JSONStorableFloat storableTonemapExposure = new JSONStorableFloat("Tonemapping exposure", initialValue, min, max, true, true);
      storableTonemapExposure.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storableTonemapExposure, true);      
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storableTonemapExposure);
      storableTonemapExposure.setCallbackFunction = (float v) => { applySetting((s) => s.setTonemappingExposure(v)); };

      tab.Add(slider);
    }

    private void initSaturationScale(List<UIDynamic> tab) {
      float initialValue = 0.0f;
      float min = -1.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Saturation", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, false);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setSaturationScale(v)); };

      tab.Add(slider);
    }

    private void initBloomControl(List<UIDynamic> tab) {
      JSONStorableBool storable = new JSONStorableBool("Bloom", true);
      storable.setCallbackFunction = (bool val) => {
        effectsCamera.setBloomEnabled(val);
      };

      RegisterBool(storable);
      var toggle = CreateToggle(storable, false);

      tab.Add(toggle);
    }

    private void initSharpeningScale(List<UIDynamic> tab) {
      float initialValue = 0.3f;
      float min = 0.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Sharpening", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, false);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setSharpeningScale(v)); };

      tab.Add(slider);
    }

    private void initVignetteScale(List<UIDynamic> tab) {
      float initialValue = 0.0f;
      float min = 0.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Vignette", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, false);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setVignetteScale(v)); };

      tab.Add(slider);
    }

    private void initChromaticAberrationScale(List<UIDynamic> tab) {
      float initialValue = 0.0f;
      float min = 0.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Chromatic Aberration", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, false);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setChromaticAberrationScale(v)); };

      tab.Add(slider);
    }

    private void initDepthOfFieldToggle(List<UIDynamic> tab) {
      JSONStorableBool storable = new JSONStorableBool("Depth of Field", false);
      storable.setCallbackFunction = (bool val) => {
        effectsCamera.setDepthOfFieldEnabled(val);
      };

      RegisterBool(storable);
      var toggle = CreateToggle(storable, true);

      tab.Add(toggle);
    }

    private void initFocalLength(List<UIDynamic> tab) {
      float initialValue = 50.0f;
      float min = 1.0f;
      float max = 300.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Focal Length (mm)", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, true);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { effectsCamera.setDepthOfFieldFocalLength(v); };

      tab.Add(slider);
    }

    private void initApertureSize(List<UIDynamic> tab) {
      float initialValue = 5.6f;
      float min = 0.05f;
      float max = 32.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Aperture (f-stop)", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, true);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { effectsCamera.setDepthOfFieldAperture(v); };

      tab.Add(slider);
    }

    private void initKernelSize(List<UIDynamic> tab) {
      List<string> kernelSizes = new List<string>();
      kernelSizes.Add("Small");
      kernelSizes.Add("Medium");
      kernelSizes.Add("Large");
      kernelSizes.Add("Very Large");

      JSONStorableStringChooser.SetStringCallback pickerCallback = (string s) => {
        switch(s) {
          case "Small":
            effectsCamera.setDepthOfFieldKernelSize(DepthOfFieldModel.KernelSize.Small);
            break;
          case "Medium":
            effectsCamera.setDepthOfFieldKernelSize(DepthOfFieldModel.KernelSize.Medium);
            break;
          case "Large":
            effectsCamera.setDepthOfFieldKernelSize(DepthOfFieldModel.KernelSize.Large);
            break;
          case "Very Large":
            effectsCamera.setDepthOfFieldKernelSize(DepthOfFieldModel.KernelSize.VeryLarge);
            break;
        }
      };
      JSONStorableStringChooser storable = new JSONStorableStringChooser("Kernel Size", kernelSizes, "Medium", "Kernel Size", pickerCallback); // pickerCallback
      UIDynamicPopup dropdown = CreatePopup(storable, true);
      RegisterStringChooser(storable);

      tab.Add(dropdown);
    }

    private void initAutoFocus(List<UIDynamic> tab) {
      float initialValue = 1.5f;
      float min = 0.1f;
      float max = 100.0f;

      UIDynamicPopup targetPopup = null;
      UIDynamicSlider distanceSlider = null;

      List<string> atomList = new List<string>();
      atomList.Add("Manual");

      List<string> targetList = new List<string>();

      JSONStorableStringChooser storableSubTarget = new JSONStorableStringChooser("Target", targetList, "head", "Target");
      RegisterStringChooser(storableSubTarget);

      JSONStorableFloat storableDistance = new JSONStorableFloat("Focus Distance", initialValue, min, max, true, true);
      storableDistance.storeType = JSONStorableParam.StoreType.Full;
      RegisterFloat(storableDistance);

      Dictionary<string, Func<GameObject>> atomDictionary = new Dictionary<string, Func<GameObject>>();
      atomDictionary.Add("Manual", () => null);

      var types = new [] { "Person" };
      var filteredAtoms = SuperController.singleton.GetAtoms().Where(atom => types.Any(atom.type.Equals));

      foreach (var atom in filteredAtoms) {
        atomDictionary.Add(atom.uid, () => null);
        atomList.Add(atom.uid);
      }  

      // OnAtomUIDsChanged maybe?

      JSONStorableStringChooser.SetStringCallback chooserCallback = (string s) => {
        switch(s) {
          case "Manual":
            effectsCamera.setAutoFocusTarget(null);
            storableSubTarget.setCallbackFunction = null;

            if (targetPopup != null) {
              tab.Remove(targetPopup);
              RemovePopup(targetPopup);
              targetPopup = null;
            }

            if (distanceSlider == null) {
              distanceSlider = CreateSlider(storableDistance, true);
              distanceSlider.valueFormat = "F3";
              distanceSlider.AutoSetButtons();

              storableDistance.setCallbackFunction = (float v) => { effectsCamera.setDepthOfFieldFocusDistance(v); };

              tab.Add(distanceSlider);
            }

            effectsCamera.setDepthOfFieldFocusDistance(storableDistance.val);
            break;
          default: 
            if (distanceSlider != null) {
              tab.Remove(distanceSlider);
              RemoveSlider(distanceSlider);
              distanceSlider = null;
            }

            var atom = SuperController.singleton.GetAtomByUid(s);
            var filteredRBs = atom.linkableRigidbodies.Where(rb => atom.forceReceivers.Select(fr => fr.name).Any(rb.name.Equals)); // filter to the non control things
            
            {
              var name = storableSubTarget.val;
              var target = filteredRBs.FirstOrDefault(rb => name == rb.name);
              effectsCamera.setAutoFocusTarget(target.gameObject);
            }

            storableSubTarget.choices = filteredRBs.Select((rb) => rb.name).ToList();

            if (targetPopup == null) {
              targetPopup = CreateFilterablePopup(storableSubTarget, true);
              tab.Add(targetPopup);
            }

            storableSubTarget.setCallbackFunction = (name) => {
              var target = filteredRBs.FirstOrDefault(rb => name == rb.name);
              
              if(target != null) {
                effectsCamera.setAutoFocusTarget(target.gameObject);
              }
            };

            break;
        }
      };

      JSONStorableStringChooser storable = new JSONStorableStringChooser("Focus", atomList, "Manual", "Focus", chooserCallback); // pickerCallback
      storable.representsAtomUid = true;
      UIDynamicPopup dropdown = CreateFilterablePopup(storable, true);
      RegisterStringChooser(storable);

      if (distanceSlider == null && storable.val == storable.defaultVal) {
        distanceSlider = CreateSlider(storableDistance, true);
        distanceSlider.valueFormat = "F3";
        distanceSlider.AutoSetButtons();

        storableDistance.setCallbackFunction = (float v) => { effectsCamera.setDepthOfFieldFocusDistance(v); };

        tab.Add(distanceSlider);
      }

      storable.popupOpenCallback = () => {
        filteredAtoms = SuperController.singleton.GetAtoms().Where(atom => types.Any(atom.type.Equals));

        var newList = new List<string> { "Manual" };
        newList.AddRange(filteredAtoms.Select((atom) => atom.name).ToList());

        storable.choices = newList;
      };


      tab.Add(dropdown);
    }

    private void initSoftenScale(List<UIDynamic> tab) {
      float initialValue = 1.0f;
      float min = 0.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Soften", initialValue, min, max, true, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, true);
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { applySetting((s) => s.setSoftenScale(v)); };

      tab.Add(slider);
    }

    private void initAtmosphereDensity(List<UIDynamic> tab) {
      float initialValue = 0.0f;
      float min = 0.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Atmosphere Density", initialValue, min, max, false, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, true);
      slider.label = "Density";
      slider.valueFormat = "F3";
      slider.rangeAdjustEnabled = true;
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { effectsCamera.setAtmosphereDensity(v); };

      tab.Add(slider);
    }

    private void initAtmosphereSeaLevel(List<UIDynamic> tab) {
      float initialValue = 0.0f;
      float min = -10.0f;
      float max = 10.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Atmosphere Sea Level", initialValue, min, max, false, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, true);
      slider.label = "Sea Level";
      slider.valueFormat = "F3";
      slider.rangeAdjustEnabled = true;
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { effectsCamera.setAtmosphereSeaLevel(v); };

      tab.Add(slider);
    }

    private void initAtmosphereSeaLevelDensity(List<UIDynamic> tab) {
      float initialValue = 0.0f;
      float min = 0.0f;
      float max = 1.0f;
      JSONStorableFloat storable = new JSONStorableFloat("Atmosphere Sea Level Density", initialValue, min, max, false, true);
      storable.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storable, true);
      slider.label = "Sea Level Density";
      slider.valueFormat = "F3";
      slider.rangeAdjustEnabled = true;
      slider.AutoSetButtons();

      RegisterFloat(storable);
      storable.setCallbackFunction = (float v) => { effectsCamera.setAtmosphereSeaLevelDensity(v); };

      tab.Add(slider);
    }

    private void initCustomLUT(List<UIDynamic> tab) {
      TextureSetCallback callback = TextureLoaded;

      List<string> luts = new List<string>();
      luts.Add("None");
      luts.Add("Moyashi Default");
      luts.Add("Moyashi Energetic");
      luts.Add("Unity Default");
      luts.Add("Custom");

      // built in LUT picker
      JSONStorableUrl storableLutUrl = new JSONStorableUrl("LUT_texture", getPackagePath() + "Custom/Images/LUT/MoyashiDefaultLUT32.png", (string url) => { QueueLoadTexture(url, callback); }, "jpg|png|tif|tiff");
      if (storableLutUrl.val == storableLutUrl.defaultVal) {
        QueueLoadTexture(storableLutUrl.val, callback);
      }
      RegisterUrl(storableLutUrl);

      JSONStorableBool storableCustomSwitch = new JSONStorableBool("LUT_custom_enabled", false);
      RegisterBool(storableCustomSwitch);

      // Check if Custom/Images/LUT/ exists, if not create it.
      if (!FileManagerSecure.DirectoryExists("Custom/Images/LUT", true)) {
        FileManagerSecure.CreateDirectory("Custom/Images/LUT");
      }

      JSONStorableUrl storableLutCustomUrl = new JSONStorableUrl("LUT_custom", string.Empty, (string url) => { 
        if (storableCustomSwitch.val) {
          storableLutUrl.val = url; 
        }
      }, "jpg|png|tif|tiff", "Custom/Images/LUT/");

      List<ShortCut> shortCuts = new List<ShortCut>();

      // hecking booleans, 1st arg: flatten or something? 2nd arg: allow browsing whole pc, 3rd arg include regular dirs
      shortCuts.AddRange(FileManagerSecure.GetShortCutsForDirectory("Custom/Images/LUT/", true, false, true)); 
      
      // GET READY FOR SOME REAL HACKS HOLY
      // basically prefills some actual useful packages in the sidebar rather than mostly poorly packaged stuff
      var creators = new [] { "ICannotDie", "Killer2u", "BooMoon"};
      var directories = new [] { "Custom/LUTs/", "Custom/Assets/ICannotDie/LUT32/", "Custom/Assets/MacGruber/PostMagic/BooMoon_LUT_PACK/" };
      foreach(string directory in directories) {
        var otherShortCuts = FileManagerSecure.GetShortCutsForDirectory(directory);
        if (otherShortCuts.Count > 0) {
          var filteredShortCuts = otherShortCuts.Where(osc => creators.Any(osc.creator.Contains)).ToList();
          shortCuts.AddRange(filteredShortCuts);
        }
      }

      shortCuts.Insert(0, new ShortCut {
        displayName = "Custom\\Images",
        path = FileManagerSecure.NormalizePath("Custom/Images"),
        // flatten = true,
      });

      shortCuts.Insert(0, new ShortCut {
        displayName = "Custom",
        path = FileManagerSecure.NormalizePath("Custom"),
        // flatten = true,
      });

      shortCuts.Insert(0, new ShortCut {
        displayName = "AddonPackages",
        path = FileManagerSecure.NormalizePath("AddonPackages"),
        // flatten = true,
      });

      shortCuts.Insert(0, new ShortCut {
        displayName = "Root",
        path = FileManagerSecure.NormalizePath("."),
        // flatten = true,
      });

      storableLutCustomUrl.beginBrowseWithObjectCallback = (JSONStorableUrl innerStorable) => {
        innerStorable.shortCuts = shortCuts;
      };

      RegisterUrl(storableLutCustomUrl);
      
      JSONStorableStringChooser.SetStringCallback pickerCallback = (string s) => {
        switch(s) {
          case "None":
            storableCustomSwitch.val = false;
            storableLutUrl.val = getPackagePath() + "Custom/Images/LUT/NeutralLUT32.png";
            break;
          case "Moyashi Energetic":
            storableCustomSwitch.val = false;
            storableLutUrl.val = getPackagePath() + "Custom/Images/LUT/MoyashiEnergeticLUT32.png";
            break;
          case "Moyashi Default":
            storableCustomSwitch.val = false;
            storableLutUrl.val = getPackagePath() + "Custom/Images/LUT/MoyashiDefaultLUT32.png";
            break;
          case "Unity Default":
            storableCustomSwitch.val = false;
            storableLutUrl.val = getPackagePath() + "Custom/Images/LUT/UnityDefaultLUT32.png";
            break;
          case "Custom":
            storableCustomSwitch.val = true;
            storableLutUrl.val = storableLutCustomUrl.val;
            break;
        }
      };

      JSONStorableStringChooser storable = new JSONStorableStringChooser("LUT_builtin", luts, "Moyashi Default", "LUT", pickerCallback); // pickerCallback
      UIDynamicPopup dropdown = CreatePopup(storable, true);
      RegisterStringChooser(storable);

      // LUT strength
      float initialValue = 1.0f;
      float min = 0.0f;
      float max = 1.0f;
      JSONStorableFloat storableLutStrength = new JSONStorableFloat("LUT strength", initialValue, min, max, true, true);
      storableLutStrength.storeType = JSONStorableParam.StoreType.Full;

      UIDynamicSlider slider = CreateSlider(storableLutStrength, true);      
      slider.valueFormat = "F3";
      slider.AutoSetButtons();

      RegisterFloat(storableLutStrength);
      storableLutStrength.setCallbackFunction = (float v) => { applySetting((s) => s.setLUTContribution(v)); };
      storableCustomSwitch.setCallbackFunction = (bool visible) => {
        if (visible && customLutButton == null && customLutTextfield == null) {
          customLutButton = CreateButton("Browse for Custom LUT", true);
          customLutTextfield = CreateTextField(storableLutCustomUrl, true);
          customLutTextfield.height = 0.0f;
          storableLutCustomUrl.RegisterFileBrowseButton(customLutButton.button);
          tab.Add(customLutButton);
          tab.Add(customLutTextfield);
        } else if (customLutButton != null && customLutTextfield != null) {
          tab.Remove(customLutButton);
          tab.Remove(customLutTextfield);
          RemoveButton(customLutButton);
          RemoveTextField(customLutTextfield);
        }
      };

      tab.Add(dropdown);
      tab.Add(slider);
    }

    private void initSkyColor(List<UIDynamic> tab) {
      Color initialSkyColor = new Color(0.7f, 0.93f, 1.0f);

      HSVColor hsvSkyColor = HSVColorPicker.RGBToHSV(initialSkyColor.r, initialSkyColor.g, initialSkyColor.b);
      JSONStorableColor skyColor = new JSONStorableColor("Sky Color", hsvSkyColor); // should this be JSONStorableVector3?
      skyColor.storeType = JSONStorableParam.StoreType.Full;
      var picker = CreateColorPicker(skyColor, false);
      RegisterColor(skyColor);
      skyColor.setCallbackFunction = (float h, float s, float v) => { effectsCamera.setAtmosphereSkyColor(HSVColorPicker.HSVToRGB(h, s, v)); };

      // hack fiesta
      Defer(() => { 
        var rect = picker.colorPicker.Selector.rectTransform;
        var container = picker.colorPicker.Container;

        if (rect != null && container != null && container.rect != null) {
          rect.anchoredPosition = new Vector2(
            255 * picker.colorPicker.xVal,
            container.rect.height * picker.colorPicker.yVal
          );
        }
      }, 0.0f);

      tab.Add(picker);
    }

    private void initSunColor(List<UIDynamic> tab) {
      Color initialSunColor = new Color(1.0f, 0.99f, 0.83f);

      HSVColor hsvSunColor = HSVColorPicker.RGBToHSV(initialSunColor.r, initialSunColor.g, initialSunColor.b);
      JSONStorableColor sunColor = new JSONStorableColor("Sun Color", hsvSunColor); // should this be JSONStorableVector3?
      sunColor.storeType = JSONStorableParam.StoreType.Full;
      var picker = CreateColorPicker(sunColor, false);
      RegisterColor(sunColor);
      sunColor.setCallbackFunction = (float h, float s, float v) => { effectsCamera.setAtmosphereSunColor(HSVColorPicker.HSVToRGB(h, s, v)); };

      // hack fiesta
      Defer(() => { 
        var rect = picker.colorPicker.Selector.rectTransform;
        var container = picker.colorPicker.Container;

        if (rect != null && container != null && container.rect != null) {
          rect.anchoredPosition = new Vector2(
            255 * picker.colorPicker.xVal,
            container.rect.height * picker.colorPicker.yVal
          );
        }
      }, 0.0f);

      tab.Add(picker);
    }

    // This runs first when the plugin is loaded, after enable, init gets run
    private void OnEnable() {
      // SuperController.LogMessage("OnEnable");
      Camera mainCamera = CameraTarget.centerTarget.targetCamera;

      // Create a special camera just for doing post processing
      effectsCamera = mainCamera.GetComponent<EffectsCamera>();
      if (effectsCamera == null) {
        effectsCamera = mainCamera.gameObject.AddComponent<EffectsCamera>();
      }

      // TODO: reimplement effect as command buffers for screenshot/thumbnail cameras
      // screenshotCamera = SuperController.singleton.hiResScreenshotCamera;
      // getOrAddCustomPostProcessingComponentToCamera(screenshotCamera);

      // thumbnailCamera = SuperController.singleton.screenshotCamera;
      // getOrAddCustomPostProcessingComponentToCamera(thumbnailCamera);



      // SuperController.LogMessage("Available Layer masks");

      // for(int i = 0; i < 32; i++) {
        // SuperController.LogMessage($" - {i} {LayerMask.LayerToName(i)}");
      // }
    }

    private void styleButtonAsTab(UIDynamicButton button, bool active) {
      var color = active ? new Color(0.42f, 0.03f, 0.51f) : new Color(0.63f, 0.54f, 0.66f);

      button.buttonColor = color;
      button.buttonText.color = Color.white;
      button.buttonText.alignment = TextAnchor.UpperCenter;
      button.buttonText.fontStyle = FontStyle.Bold;
      {
        var offsetMax = button.buttonText.rectTransform.offsetMax;
        offsetMax.y = -10.5f;
        button.buttonText.rectTransform.offsetMax = offsetMax;
      }
    }

    private void setActive(List<UIDynamic> elements) {
      Defer(() => {
        foreach(var element in elements) {
          element.gameObject.SetActive(true);
        }
      }, 0.0f);
    }

    private void setInactive(List<UIDynamic> elements) {
      Defer(() => {
        foreach(var element in elements) {
          element.gameObject.SetActive(false);
        }
      }, 0.0f);
    }

    private void handleTabClick(UIDynamicButton triggeringTab) {
      foreach(var tab in tabs) {
        styleButtonAsTab(tab, false);
      }

      styleButtonAsTab(triggeringTab, true);

      switch(triggeringTab.label) {
        case "Color Grading":
          setActive(colorGradingTab);
          setInactive(lensTab);
          setInactive(atmosphereTab);
          break;
        case "Lens":
          setInactive(colorGradingTab);
          setActive(lensTab);
          setInactive(atmosphereTab);
          break;
        case "Atmosphere":
          setInactive(colorGradingTab);
          setInactive(lensTab);
          setActive(atmosphereTab);
          break;
      }
    }

    private void initTabs() {
      float height = 40.0f;

      UIDynamic spacerL = CreateSpacer(false);
      spacerL.height = height;

      UIDynamic spacerR = CreateSpacer(true);
      spacerR.height = height;

      UIDynamicButton colorCorrectionTab = CreateButton("Color Grading", false);
      styleButtonAsTab(colorCorrectionTab, true);
      tabs.Add(colorCorrectionTab);
      colorCorrectionTab.button.onClick.AddListener(() => handleTabClick(colorCorrectionTab));

      Defer(() => {
        colorCorrectionTab.transform.parent = spacerL.transform;
        LayoutElement layout = colorCorrectionTab.GetComponent<LayoutElement>();
        layout.ignoreLayout = true;
        RectTransform rect = colorCorrectionTab.GetComponent<RectTransform>();
        rect.sizeDelta = new Vector2(0.0f, 0.0f);
        rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 0.0f, 210.0f); 
        rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 0.0f, 60.0f);
      }, 0.1f);

      UIDynamicButton lensTab = CreateButton("Lens", false);
      styleButtonAsTab(lensTab, false);
      tabs.Add(lensTab);
      lensTab.button.onClick.AddListener(() => handleTabClick(lensTab));

      Defer(() => {
        lensTab.transform.parent = spacerL.transform;
        LayoutElement layout = lensTab.GetComponent<LayoutElement>();
        layout.ignoreLayout = true;
        RectTransform rect = lensTab.GetComponent<RectTransform>();
        rect.sizeDelta = new Vector2(0.0f, 0.0f);
        rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 215.0f, 85.0f); 
        rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 0.0f, 60.0f);
      }, 0.1f);

      UIDynamicButton atmosphereTab = CreateButton("Atmosphere", false);
      styleButtonAsTab(atmosphereTab, false);
      tabs.Add(atmosphereTab);
      atmosphereTab.button.onClick.AddListener(() => handleTabClick(atmosphereTab));

      Defer(() => {
        atmosphereTab.transform.parent = spacerL.transform;
        LayoutElement layout = atmosphereTab.GetComponent<LayoutElement>();
        layout.ignoreLayout = true;
        RectTransform rect = atmosphereTab.GetComponent<RectTransform>();
        rect.sizeDelta = new Vector2(0.0f, 0.0f);
        rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 305.0f, 185.0f); 
        rect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 0.0f, 60.0f);
      }, 0.1f);    
    }

    // UIDynamicTextField textfield = CreateTextField(storable, alignRight);
    // textfield.textColor = new Color(1,1,1);
    // textfield.height = 30.0f;
    // textfield.UItext.transform.parent = mixer;
    // textfield.UItext.fontSize = 24;
    // textfield.UItext.fontStyle = FontStyle.Bold;
    // textfield.UItext.alignment = TextAnchor.MiddleCenter;

    // LayoutElement component = textfield.GetComponent<LayoutElement>();
    // component.ignoreLayout = true;

    // RectTransform textfieldRect = textfield.GetComponent<RectTransform>();
    // textfieldRect.sizeDelta = new Vector2(0.0f, 0.0f);

    // RectTransform textRect = textfield.UItext.GetComponent<RectTransform>();
    // textRect.localPosition = new Vector3(0,0,0);
    // textRect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Top, 2.5f, 30.0f);
    // textRect.SetInsetAndSizeFromParentEdge(RectTransform.Edge.Left, 10.0f, 500.0f);







    // init happens after onenable
    public override void Init() {
      // SuperController.LogMessage("Init");
      // this is roughly in order of what the shader does, whether thats helpful or not

      initTabs();

      initSaturationScale(colorGradingTab);
      initContrastScale(colorGradingTab);
      initChannelMixer(colorGradingTab); 
      initSoftenScale(colorGradingTab);
      initTonemappingSelect(colorGradingTab);
      initTonemappingExposure(colorGradingTab);
      initCustomLUT(colorGradingTab);

      initBloomControl(lensTab);
      initSharpeningScale(lensTab);
      initVignetteScale(lensTab);
      initChromaticAberrationScale(lensTab);
      initDepthOfFieldToggle(lensTab);
      initKernelSize(lensTab);
      initApertureSize(lensTab);
      initFocalLength(lensTab);
      initAutoFocus(lensTab);

      initSkyColor(atmosphereTab);
      initSunColor(atmosphereTab);
      initAtmosphereDensity(atmosphereTab);
      initAtmosphereSeaLevelDensity(atmosphereTab);
      initAtmosphereSeaLevel(atmosphereTab);

      handleTabClick(tabs.First());

      effectsCamera.enabled = true;

      // TODO: i think initializing the rest of the UI needs to wait for this shader to load first...
      // but eh, just work around it
      StartCoroutine("loadShaders");
    }

    private void OnDisable() {
      // SuperController.LogMessage("main OnDisable");
      
    }

    // This runs when either the plugin is removed or is reloaded
    private void OnDestroy() {
      // SuperController.LogMessage("main OnDisable");
      StopCoroutine("loadShaders");

      if (texture != null) {
        DestroyImmediate(texture);
      }

      texture = null;   

      {
        // CustomPostProcessing cpp = screenshotCamera.GetComponent<CustomPostProcessing>();
        // DestroyImmediate(cpp);
        // SuperController.LogMessage("screenshotCamera custom post processing destroyed");
      }

      {
        // CustomPostProcessing cpp = thumbnailCamera.GetComponent<CustomPostProcessing>();
        // DestroyImmediate(cpp);
        // SuperController.LogMessage("thumbnailCamera custom post processing destroyed");
      }

      // customPostProcessingSet.Clear();  

      effectsCamera.enabled = false;
      DestroyImmediate(effectsCamera);

      // make sure any queued image loads don't run callbacks after destroy
      queuedImage.callback = null;

      AssetLoader.DoneWithAssetBundleFromFile(getPackagePath() + "Custom/Assets/moyashi/Moyashi_CustomColorCorrection.assetbundle");
      // SuperController.LogMessage("main OnDestroy complete");
    }

    private void TextureLoaded(Texture2D tex) {           
      if (texture != null) {
        Destroy(texture);
      }

      if (tex != null) {
        tex.filterMode = FilterMode.Bilinear;

        // CustomPostProcessing customPostProcessing = effectsCamera.getOrAddCustomPostProcessingComponent();
        // customPostProcessing.setLUTTexture(tex);

        effectsCamera.setLUTTexture(tex);

        // foreach(CustomPostProcessing cpp in customPostProcessingSet) {
        //   cpp.setLUTTexture(tex);
        // }
      }

      queuedImage = null;
    }

    private void QueueLoadTexture(string url, TextureSetCallback callback) { 
      if (ImageLoaderThreaded.singleton == null) {
        SuperController.LogError("no ImageLoaderThreaded, old version of vam?");
        return;
      }

      if (string.IsNullOrEmpty(url)) {
        // This happens when selecting "custom" for the first time...
        // so i guess not really an error just handle it silently
        // SuperController.LogError("null or empty url");
        return;
      }

      // prevent the callbacks stepping on each others toes
      if (queuedImage != null) {
        queuedImage.callback = null;
      }

      queuedImage = new ImageLoaderThreaded.QueuedImage();
      queuedImage.compress = false;
      queuedImage.createMipMaps = false;
      queuedImage.forceReload = true;
      queuedImage.imgPath = url;
      queuedImage.linear = true;
      queuedImage.callback = (ImageLoaderThreaded.QueuedImage queuedImage) => {
        if (queuedImage.tex != null) {
          queuedImage.tex.wrapMode = TextureWrapMode.Clamp;
          queuedImage.tex.filterMode = FilterMode.Bilinear;
          queuedImage.tex.anisoLevel = 0;
          callback(queuedImage.tex);
        }        
      };

      ImageLoaderThreaded.singleton.QueueImage(queuedImage);
    }
  }

  public delegate void TextureSetCallback(Texture2D tex);

  public class SharedUniforms {
    // need to temp store uniforms here in case they get set before the shader is loaded... kinda annoying
    // but avoids race condition between shader and stored json stuff loading

    // color correction
    public Vector3 ChannelMixR = new Vector3(1.0f, 0.0f, 0.0f);
    public Vector3 ChannelMixG = new Vector3(0.0f, 1.0f, 0.0f);
    public Vector3 ChannelMixB = new Vector3(0.0f, 0.0f, 1.0f);
    public float TonemappingScale = 1.0f;
    public float TonemappingExposure = 1.0f;
    public float ContrastScale = 0.0f;
    public float SaturationScale = 0.0f;
    public float SoftenScale = 1.0f;
    public float AtmosphereDensity = 0.0f;
    public float AtmosphereSeaLevel = 0.0f;
    public float AtmosphereSeaLevelDensity = 0.0f;
    public Color AtmosphereSkyColor = new Color(0.7f, 0.93f, 1.0f, 1.0f);
    public Color AtmosphereSunColor = new Color(1.0f, 0.99f, 0.83f, 1.0f);
    public float LUTScale = 1.0f;
    public Texture LUTTex;

    // fidelity fx CAS
    public float SharpeningScale = 0.3f;
    public Vector4 SharpeningConst0 = new Vector4(0.0f, 0.0f, 0.0f, 0.0f);
    public Vector4 SharpeningConst1 = new Vector4(0.0f, 0.0f, 0.0f, 0.0f);
    public float VignetteScale = 0.0f;
    public float ChromaticAberrationScale = 0.0f;
  }

  public class PrepassUniforms {
    public float AtmosphereDensity = 0.0f;
    public float AtmosphereSeaLevel = 0.0f;
    public float AtmosphereSeaLevelDensity = 0.0f;
    public Color AtmosphereSkyColor = new Color(0.7f, 0.93f, 1.0f, 1.0f);
    public Color AtmosphereSunColor = new Color(1.0f, 0.99f, 0.83f, 1.0f);
  }

  public struct CSMeta {
    public int kernelIndex;
    public uint threadGroupsX;
    public uint threadGroupsY;
    public uint threadGroupsZ;
  }

  public class UIContainer : Graphic {
    protected override void OnPopulateMesh(VertexHelper vh) {
      Vector2 corner1 = Vector2.zero;
      Vector2 corner2 = Vector2.zero;

      corner1.x = 0f;
      corner1.y = 0f;
      corner2.x = 1f;
      corner2.y = 1f;

      corner1.x -= rectTransform.pivot.x;
      corner1.y -= rectTransform.pivot.y;
      corner2.x -= rectTransform.pivot.x;
      corner2.y -= rectTransform.pivot.y;

      corner1.x *= rectTransform.rect.width;
      corner1.y *= rectTransform.rect.height;
      corner2.x *= rectTransform.rect.width;
      corner2.y *= rectTransform.rect.height;

      vh.Clear();

      UIVertex vert = UIVertex.simpleVert;

      vert.position = new Vector2(corner1.x, corner1.y);
      vert.color = new Color(1.0f, 1.0f, 1.0f, 0.2f);
      vh.AddVert(vert);

      vert.position = new Vector2(corner1.x, corner2.y);
      vert.color = new Color(1.0f, 1.0f, 1.0f, 0.2f);
      vh.AddVert(vert);

      vert.position = new Vector2(corner2.x, corner2.y);
      vert.color = new Color(1.0f, 1.0f, 1.0f, 0.2f);
      vh.AddVert(vert);

      vert.position = new Vector2(corner2.x, corner1.y);
      vert.color = new Color(1.0f, 1.0f, 1.0f, 0.2f);
      vh.AddVert(vert);

      vh.AddTriangle(0, 1, 2);
      vh.AddTriangle(2, 3, 0);
    }
  }

  [System.Runtime.InteropServices.StructLayout(System.Runtime.InteropServices.LayoutKind.Explicit)]
  public struct f32uintUnion {
    [System.Runtime.InteropServices.FieldOffset(0)]
    public float f;

    [System.Runtime.InteropServices.FieldOffset(0)]
    public uint u;
  }

  public class EffectsCamera : MonoBehaviour {
    private static int DEFAULT_CULLING_MASK = -142606337;

    private Camera postProcessingCamera;
    private Camera mainCamera;
    private int uiMask = 0;
    private int worldMask = 0;
    
    private int originalMainCullingMask = -1;
    private float originalMainDepth = 0.0f;
  
    private BloomModel.Settings bloomSettings;
    private DepthOfFieldModel.Settings depthOfFieldSettings;

    private GameObject autoFocusTarget = null;

    private PostProcessingBehaviour postProcessingBehaviour;
    public PostProcessingProfile profile { get; private set; }

    private PrepassUniforms prepassUniforms;
    private SharedUniforms uniforms;

    // pre unity post proc pass
    private Shader prepassShader;
    private Material prepassMaterial;
    private CommandBuffer prepassCommandBuffer;

    // post pass
    private Material colorCorrectionMaterial;
    private ComputeShader compute;
    private ComputeShader lensCompute;
    private CSMeta computeMeta;
    private CSMeta lensComputeMeta;
    private CommandBuffer commandBuffer;

    // copy to main cam
    private CommandBuffer finalCommandBuffer;

    private RenderTexture renderTexture;

    // https://docs.unity3d.com/2018.1/Documentation/ScriptReference/Rendering.CameraEvent.html
    private CameraEvent PRE_PROC_EVENT = CameraEvent.BeforeForwardAlpha;
    private CameraEvent POST_PROC_EVENT = CameraEvent.BeforeImageEffects;

    public void OnPreRender() {
      postProcessingCamera.targetTexture = renderTexture;
    }

    public void OnPostRender() {
      postProcessingCamera.targetTexture = null;
    }

    public void setPrepassShader(Shader shader) {
      // getOrAddCustomPostProcessingPrepassComponent().setPrepassShader(shader);
      prepassShader = shader;
      prepassMaterial = new Material(shader);

      prepassCommandBuffer = new CommandBuffer();
      prepassCommandBuffer.name = "Moyashi post processing prepass";

      int screenCopyId = Shader.PropertyToID("_MoyashiPrePassRT");
      // SuperController.LogMessage($"id: {screenCopyId}");
      prepassCommandBuffer.GetTemporaryRT(screenCopyId, -1, -1, 0, FilterMode.Bilinear);
      // prepassCommandBuffer.SetRenderTarget(BuiltinRenderTextureType.CameraTarget);
      prepassCommandBuffer.Blit(BuiltinRenderTextureType.CameraTarget, screenCopyId, this.prepassMaterial);

      // prepassCommandBuffer.ClearRenderTarget(true, true, Color.red, 1.0f);
      // prepassCommandBuffer.Blit(BuiltinRenderTextureType.CurrentActive, BuiltinRenderTextureType.CameraTarget, this.prepassMaterial);
      // prepassCommandBuffer.ClearRenderTarget(true, true, Color.black);
      prepassCommandBuffer.Blit(screenCopyId, BuiltinRenderTextureType.CameraTarget);

      prepassCommandBuffer.ReleaseTemporaryRT(screenCopyId);
      // prepassCommandBuffer.SetRenderTarget(BuiltinRenderTextureType.CameraTarget, BuiltinRenderTextureType.Depth);
      // postProcessingCamera.AddCommandBuffer(PRE_PROC_EVENT, prepassCommandBuffer);
      // SuperController.LogMessage("set command buffer");

      updateUniforms();
    }

    public void setShaders(Shader colorCorrectionShader, ComputeShader compute, ComputeShader lensCompute) {
      colorCorrectionMaterial = new Material(colorCorrectionShader);
      this.compute = compute;
      this.lensCompute = lensCompute;

      this.computeMeta.kernelIndex = this.compute.FindKernel("MoyashiCompute");

      this.compute.GetKernelThreadGroupSizes(this.computeMeta.kernelIndex,
        out this.computeMeta.threadGroupsX,
        out this.computeMeta.threadGroupsY,
        out this.computeMeta.threadGroupsZ
      );

      this.lensComputeMeta.kernelIndex = this.lensCompute.FindKernel("MoyashiLens");

      this.lensCompute.GetKernelThreadGroupSizes(this.lensComputeMeta.kernelIndex,
        out this.lensComputeMeta.threadGroupsX,
        out this.lensComputeMeta.threadGroupsY,
        out this.lensComputeMeta.threadGroupsZ
      );

      commandBuffer = new CommandBuffer();
      commandBuffer.name = "Moyashi post processing";

      int pingRT = Shader.PropertyToID("_MoyashiPostProcessingPingRT");
      commandBuffer.GetTemporaryRT(pingRT, -1, -1, 0, FilterMode.Bilinear, RenderTextureFormat.ARGB32, RenderTextureReadWrite.Default, 1, true);

      int pongRT = Shader.PropertyToID("_MoyashiPostProcessingPongRT");
      commandBuffer.GetTemporaryRT(pongRT, -1, -1, 0, FilterMode.Bilinear, RenderTextureFormat.ARGB32, RenderTextureReadWrite.Default, 1, true);

      commandBuffer.Blit(BuiltinRenderTextureType.CameraTarget, pongRT, this.prepassMaterial);
      commandBuffer.Blit(pongRT, pingRT, this.colorCorrectionMaterial);
      // commandBuffer.CopyTexture(BuiltinRenderTextureType.CameraTarget, pingRT);

      commandBuffer.SetComputeTextureParam(compute, computeMeta.kernelIndex, "Read", pingRT);
      commandBuffer.SetComputeTextureParam(compute, computeMeta.kernelIndex, "Result", pongRT);
      commandBuffer.DispatchCompute(compute, computeMeta.kernelIndex, 
        postProcessingCamera.pixelWidth/(int)computeMeta.threadGroupsX, 
        postProcessingCamera.pixelHeight/(int)computeMeta.threadGroupsY, 
        (int)computeMeta.threadGroupsZ
      );

      commandBuffer.SetComputeTextureParam(lensCompute, lensComputeMeta.kernelIndex, "Read", pongRT);
      commandBuffer.SetComputeTextureParam(lensCompute, lensComputeMeta.kernelIndex, "Result", pingRT);
      commandBuffer.DispatchCompute(lensCompute, lensComputeMeta.kernelIndex, 
        postProcessingCamera.pixelWidth/(int)lensComputeMeta.threadGroupsX, 
        postProcessingCamera.pixelHeight/(int)lensComputeMeta.threadGroupsY, 
        (int)lensComputeMeta.threadGroupsZ
      );

      commandBuffer.Blit(pingRT, BuiltinRenderTextureType.CameraTarget);
      // commandBuffer.Blit(renderTexture, BuiltinRenderTextureType.None);
      // commandBuffer.Blit(pingRT, BuiltinRenderTextureType.CameraTarget);
      // commandBuffer.CopyTexture(pingRT, BuiltinRenderTextureType.CameraTarget);
      // commandBuffer.Blit(renderTexture, null as RenderTexture);
      commandBuffer.ReleaseTemporaryRT(pingRT);
      commandBuffer.ReleaseTemporaryRT(pongRT);

      postProcessingCamera.AddCommandBuffer(POST_PROC_EVENT, commandBuffer);

      updateUniforms();
    }

    // public void setColorCorrectionShader(Shader shader) {
    //   getOrAddCustomPostProcessingComponent().setColorCorrectionShader(shader);
    // }

    // public void setPassthroughShader(Shader shader) {
    //   getOrAddCustomPostProcessingComponent().setPassthroughShader(shader);
    // }

    // public void setComputeShader(ComputeShader shader) {
    //   getOrAddCustomPostProcessingComponent().setComputeShader(shader);
    // }

    // public void setLensComputeShader(ComputeShader shader) {
    //   getOrAddCustomPostProcessingComponent().setLensComputeShader(shader);
    // }

    // TODO: remove this once bloom is in compute shader pipeline
    public void setBloomEnabled(bool b) {
      // bloomSettings.enabled = b;
      // profile.bloomSettings.settings = bloomSettings;
      profile.bloom.enabled = b;
    }

    public void setDepthOfFieldEnabled(bool b) {
      if (!postProcessingCamera.stereoEnabled) {
        profile.depthOfField.enabled = b;
        profile.depthOfField.settings = depthOfFieldSettings;
      }
    }

    public void setDepthOfFieldFocusDistance(float f) {
      depthOfFieldSettings.focusDistance = f;
      profile.depthOfField.settings = depthOfFieldSettings;
    }

    public void setDepthOfFieldAperture(float f) {
      depthOfFieldSettings.aperture = f;
      profile.depthOfField.settings = depthOfFieldSettings;
    }

    public void setDepthOfFieldFocalLength(float f) {
      depthOfFieldSettings.focalLength = f;
      profile.depthOfField.settings = depthOfFieldSettings;
    }

    public void setDepthOfFieldKernelSize(DepthOfFieldModel.KernelSize e) {
      depthOfFieldSettings.kernelSize = e;
      profile.depthOfField.settings = depthOfFieldSettings;
    }

    public void setAutoFocusTarget(GameObject go) {
      autoFocusTarget = go;
    }

    public void updateUniforms() {
      if (prepassMaterial != null) {
        prepassMaterial.SetFloat("_AtmosphereDensity", prepassUniforms.AtmosphereDensity);
        prepassMaterial.SetFloat("_AtmosphereSeaLevel", prepassUniforms.AtmosphereSeaLevel);
        prepassMaterial.SetFloat("_AtmosphereSeaLevelDensity", prepassUniforms.AtmosphereSeaLevelDensity);
        prepassMaterial.SetColor("_AtmosphereSkyColor", prepassUniforms.AtmosphereSkyColor);
        prepassMaterial.SetColor("_AtmosphereSunColor", prepassUniforms.AtmosphereSunColor);
      }

      if (colorCorrectionMaterial != null) {
        colorCorrectionMaterial.SetVector("_ChannelMixR", uniforms.ChannelMixR);
        colorCorrectionMaterial.SetVector("_ChannelMixG", uniforms.ChannelMixG);
        colorCorrectionMaterial.SetVector("_ChannelMixB", uniforms.ChannelMixB);

        colorCorrectionMaterial.SetFloat("_TonemappingScale", uniforms.TonemappingScale);
        colorCorrectionMaterial.SetFloat("_TonemappingExposure", uniforms.TonemappingExposure);
        colorCorrectionMaterial.SetFloat("_ContrastScale", uniforms.ContrastScale);
        colorCorrectionMaterial.SetFloat("_SaturationScale", uniforms.SaturationScale);
        colorCorrectionMaterial.SetFloat("_SoftenScale", uniforms.SoftenScale);
        colorCorrectionMaterial.SetFloat("_LUTScale", uniforms.LUTScale);

        if (uniforms.LUTTex != null) colorCorrectionMaterial.SetTexture("_LUTTex", uniforms.LUTTex);
      }

      if (compute != null) {
        recalculateSharpeningConsts();
        compute.SetFloat("_SharpeningScale", uniforms.SharpeningScale);
        compute.SetVector("_SharpeningConst0", uniforms.SharpeningConst0);
        compute.SetVector("_SharpeningConst1", uniforms.SharpeningConst1);
        
      }

      if (lensCompute != null) {
        lensCompute.SetFloat("_VignetteScale", uniforms.VignetteScale);
        lensCompute.SetFloat("_ChromaticAberrationScale", uniforms.ChromaticAberrationScale);
      }
    }

    public void setAtmosphereDensity(float v) {
      prepassUniforms.AtmosphereDensity = v;
      updateUniforms();
    }

    public void setAtmosphereSeaLevel(float v) {
      prepassUniforms.AtmosphereSeaLevel = v;
      updateUniforms();
    }

    public void setAtmosphereSeaLevelDensity(float v) {
      prepassUniforms.AtmosphereSeaLevelDensity = v;
      updateUniforms();
    }

    public void setAtmosphereSkyColor(Color c) {
      prepassUniforms.AtmosphereSkyColor = c;
      updateUniforms();
    }

    public void setAtmosphereSunColor(Color c) {
      prepassUniforms.AtmosphereSunColor = c;
      updateUniforms();
    }

    public void setLUTTexture(Texture2D t) {
      uniforms.LUTTex = t;
      updateUniforms();
    }

    public void setChannelMixR(Vector3 c) {
      if (c != null) {
        uniforms.ChannelMixR = c;
        updateUniforms();
      }      
    }

    public void setChannelMixG(Vector3 c) {
      if (c != null) {
        uniforms.ChannelMixG = c;
        updateUniforms();
      }    
    }

    public void setChannelMixB(Vector3 c) {
      if (c != null) {
        uniforms.ChannelMixB = c;
        updateUniforms();
      }    
    }

    public void setContrastScale(float v) {
      uniforms.ContrastScale = v;
      updateUniforms();
    }

    public void setTonemappingScale(float v) {
      uniforms.TonemappingScale = v;
      updateUniforms();
    }

    public void setTonemappingExposure(float v) {
      uniforms.TonemappingExposure = v;
      updateUniforms();
    }

    public void setSaturationScale(float v) {
      uniforms.SaturationScale = v;
      updateUniforms();
    }

    public void setSoftenScale(float v) {
      uniforms.SoftenScale = v;
      updateUniforms();
    }

    public void setLUTContribution(float v) {
      uniforms.LUTScale = v;
      updateUniforms();
    }

    public void setSharpeningScale(float v) {
      uniforms.SharpeningScale = v;
      updateUniforms();
    }

    public void setVignetteScale(float v) {
      uniforms.VignetteScale = v;
      updateUniforms();
    }

    public void setChromaticAberrationScale(float v) {
      uniforms.ChromaticAberrationScale = v;
      updateUniforms();
    }

    private void OnEnable() {       
      // SuperController.LogMessage("EffectsCamera.OnEnable running");
      uniforms = new SharedUniforms();
      prepassUniforms = new PrepassUniforms();

      uiMask = 0 
             | 1 << LayerMask.NameToLayer("UI")
             | 1 << LayerMask.NameToLayer("LoadUI")
             | 1 << LayerMask.NameToLayer("ScreenUI")
             | 1 << LayerMask.NameToLayer("GUI")
             | 1 << LayerMask.NameToLayer("ControlColliders") // this is the "toggle targets" ui
      ;
      
      // worldMask = 0
      //           | 1 << LayerMask.NameToLayer("Default") // person models etc
      //           | 1 << LayerMask.NameToLayer("DefaultAndPlayerBlock") // this seems to be some CUAs?
      //           | 1 << LayerMask.NameToLayer("Water")
      //           | 1 << LayerMask.NameToLayer("Glow")
      //           | 1 << LayerMask.NameToLayer("TransparentFX")
      //           | 1 << LayerMask.NameToLayer("Select") // why is water in this layer too?????????

                // | 1 << LayerMask.NameToLayer("ControlColliders") // this is the "toggle targets" ui
                // | 1 << LayerMask.NameToLayer("PlayerBlock") // no idea
                // | 1 << LayerMask.NameToLayer("DefaultAndPlayerBlock") // no idea
                // | 1 << LayerMask.NameToLayer("Tongue")
                // | 1 << LayerMask.NameToLayer("TongueOnly")
                // | 1 << LayerMask.NameToLayer("PhysicsMesh1")
                // | 1 << LayerMask.NameToLayer("PhysicsMesh2")
                // | 1 << LayerMask.NameToLayer("WaterSurface")
                // | 1 << LayerMask.NameToLayer("Invisible")
                // | 1 << LayerMask.NameToLayer("LookAtTrigger")
                // | 1 << LayerMask.NameToLayer("Glow") // this displays selectable "glowable" things, so i guess render stuff to glow layer to make it glow
                // | 1 << LayerMask.NameToLayer("LoadUI")
                // | 1 << LayerMask.NameToLayer("ScreenUI")
                // | 1 << LayerMask.NameToLayer("GUI")
                // | 1 << LayerMask.NameToLayer("Ignore CharacterCore")
                // | 1 << LayerMask.NameToLayer("CharacterExtremities")
                // | 1 << LayerMask.NameToLayer("Player")
                // | 1 << LayerMask.NameToLayer("1Hair")
                // | 1 << LayerMask.NameToLayer("Character")
                // | 1 << LayerMask.NameToLayer("Ignore Collision")
                // | 1 << LayerMask.NameToLayer("Ignore Character")

                // | 1 << LayerMask.NameToLayer("PhysicsMesh1")
                // | 1 << LayerMask.NameToLayer("PhysicsMesh2")
                // | 1 << LayerMask.NameToLayer("PlayerBlock")
                // | 1 << LayerMask.NameToLayer("UI")
      ;

      // apparently some cuas just use random layers so... just render all except known UI ones
      worldMask = DEFAULT_CULLING_MASK & ~uiMask;

      // for (int i = 0; i < 32; i++) {
      //   SuperController.LogMessage($"layer {i} - {LayerMask.LayerToName(i)}");
      // }

      mainCamera = CameraTarget.centerTarget.targetCamera;
      originalMainCullingMask = mainCamera.cullingMask;
      originalMainDepth = mainCamera.depth;
      mainCamera.cullingMask = uiMask;

      GameObject cameraContainer = GameObject.Find("Moyashi_EffectsCamera_CameraContainer");
      if (cameraContainer == null) {
        cameraContainer = new GameObject("Moyashi_EffectsCamera_CameraContainer");
      }

      postProcessingCamera = cameraContainer.GetComponent<Camera>();

      if (postProcessingCamera == null) {
        postProcessingCamera = cameraContainer.AddComponent<Camera>();
      }
      
      postProcessingCamera.CopyFrom(mainCamera);

      // works perfectly in vr (in conjunction with setting culling matrix)
      if (postProcessingCamera.stereoEnabled) {
        postProcessingCamera.transform.parent = SuperController.singleton.heightAdjustTransform;
      } else {
        postProcessingCamera.transform.parent = mainCamera.transform;
      }

      renderTexture = new RenderTexture(postProcessingCamera.pixelWidth, postProcessingCamera.pixelHeight, 24);
      // renderTexture = new RenderTexture(BuiltinRenderTextureType.CameraTarget);

      // definitely needed
      postProcessingCamera.cullingMask = worldMask;
      postProcessingCamera.depth = mainCamera.depth - 1;
      postProcessingCamera.enabled = true;
      postProcessingCamera.allowHDR = false;
      // think this is necessary for the screen space atmospherics... otherwise it relys on user having at least 1 directional light
      postProcessingCamera.depthTextureMode = DepthTextureMode.Depth; // this does have a fairly big performance hit
      // postProcessingCamera.depthTextureMode = DepthTextureMode.None;

      // This does improve performance but nfi how to get it rndering to backbuffer when done
      // postProcessingCamera.targetTexture = renderTexture;
      // postProcessingCamera.targetTexture = null;

      // postProcessingCamera.forceIntoRenderTexture = true;
      // postProcessingCamera.renderingPath = RenderingPath.DeferredShading;

      // these maybe not ncessary after all
      // postProcessingCamera.clearFlags = CameraClearFlags.Depth;
      // postProcessingCamera.forceIntoRenderTexture = true;
      // postProcessingCamera.clearStencilAfterLightingPass = true;
      // postProcessingCamera.renderingPath = RenderingPath.Forward;

      // these were unused, maybe worth messing around with at some point (especially when trying to get AO working)
      // postProcessingCamera.renderingPath = RenderingPath.VertexLit;
      // postProcessingCamera.renderingPath = RenderingPath.DeferredLighting;
      // postProcessingCamera.clearFlags = CameraClearFlags.Nothing;
      // postProcessingCamera.clearFlags = CameraClearFlags.SolidColor;
      // postProcessingCamera.depth = mainCamera.depth;
      // postProcessingCamera.depthTextureMode = DepthTextureMode.None;
      // postProcessingCamera.depthTextureMode = DepthTextureMode.DepthNormals;

      mainCamera.clearFlags = CameraClearFlags.Nothing;

      if (true) {
        // try to get the post processing behaviour of the postProcessingCamera
        postProcessingBehaviour = postProcessingCamera.GetComponent<PostProcessingBehaviour>();

        // if it doesnt exist, create it
        if (postProcessingBehaviour == null) {
          postProcessingBehaviour = postProcessingCamera.gameObject.AddComponent<PostProcessingBehaviour>();
        } 

        profile = postProcessingBehaviour.profile;

        // i don't know if this can ever happen?
        if (profile == null) {
          postProcessingBehaviour.profile = profile = ScriptableObject.CreateInstance<PostProcessingProfile>(); 
        } 

        postProcessingBehaviour.enabled = true; // lose about 10 fps here

        // bloom, TODO: replace with MKGlow if I can figure out the memory leak, then dont need unity
        // post processor at all
        bloomSettings = BloomModel.Settings.defaultSettings;
        bloomSettings.bloom.intensity = 0.2f;
        bloomSettings.bloom.threshold = 0.75f;
        bloomSettings.bloom.softKnee = 0.25f;
        bloomSettings.bloom.radius = 3.0f;
        bloomSettings.bloom.antiFlicker = true;

        profile.bloom.enabled = true;
        profile.bloom.settings = bloomSettings;

        // temporary DoF
        if (!postProcessingCamera.stereoEnabled) {
          depthOfFieldSettings = DepthOfFieldModel.Settings.defaultSettings;
          depthOfFieldSettings.focusDistance = 1.5f;
          depthOfFieldSettings.aperture = 5.6f;
          depthOfFieldSettings.focalLength = 50.0f;
          depthOfFieldSettings.kernelSize = DepthOfFieldModel.KernelSize.Medium;

          // profile.depthOfField.enabled = true;
          // profile.depthOfField.settings = depthOfFieldSettings;
        }
      }

      finalCommandBuffer = new CommandBuffer();
      finalCommandBuffer.name = "Moyashi copy to camera";
      finalCommandBuffer.Blit(renderTexture, BuiltinRenderTextureType.CameraTarget);
      // mainCamera.AddCommandBuffer(CameraEvent.AfterImageEffects, finalCommandBuffer);
    }

    private void Update() {
      if (!postProcessingCamera.stereoEnabled) {
        postProcessingCamera.fieldOfView = mainCamera.fieldOfView;

        if (autoFocusTarget) {
          var a = depthOfFieldSettings.focusDistance;
          var b = Vector3.Distance(mainCamera.transform.position, autoFocusTarget.transform.position);
          setDepthOfFieldFocusDistance(Mathf.Lerp(a, b, 5.0f * Time.deltaTime));
        }
      }

      // absolute MESS
      // TODO: look at SteamVRFrustumAdjust in disassembly

      // could be suffering from this issue: https://github.com/ValveSoftware/openvr/issues/1052
      if (postProcessingCamera.stereoEnabled) {
        // postProcessingCamera.ResetStereoProjectionMatrices();
        // postProcessingCamera.ResetStereoViewMatrices();

        // TODO: maybe false here should be true cause we are rendering to a render texture... could also be mistake in shader cause of this
        Matrix4x4 leftProjection = Matrix4x4.identity;
        Matrix4x4 rightProjection = Matrix4x4.identity;

        leftProjection = postProcessingCamera.GetStereoProjectionMatrix(Camera.StereoscopicEye.Left);
        rightProjection = postProcessingCamera.GetStereoProjectionMatrix(Camera.StereoscopicEye.Right);

        // if (SuperController.singleton.isOpenVR) {
        //   leftProjection = SuperController.singleton.ViveCenterCamera.GetStereoProjectionMatrix(Camera.StereoscopicEye.Left);
        //   rightProjection = SuperController.singleton.ViveCenterCamera.GetStereoProjectionMatrix(Camera.StereoscopicEye.Right);
        // } else if (SuperController.singleton.isOVR) {
        //   leftProjection = SuperController.singleton.OVRCenterCamera.GetStereoProjectionMatrix(Camera.StereoscopicEye.Left);
        //   rightProjection = SuperController.singleton.OVRCenterCamera.GetStereoProjectionMatrix(Camera.StereoscopicEye.Right);
        // } else {
        //   SuperController.LogError("projection - yet another vr camera");
        // }

        // Matrix4x4[] inverseProjection = {
        //   GL.GetGPUProjectionMatrix(leftProjection, true).inverse,
        //   GL.GetGPUProjectionMatrix(rightProjection, true).inverse,
        // };

        // // apply the fix from https://forum.unity.com/threads/solved-reconstruct-world-position-from-depth-texture-in-single-pass-stereo-vr.480003/
        // inverseProjection[0][1, 1] *= - 1;
        // inverseProjection[1][1, 1] *= - 1;

        // // TODO: try this, ref https://docs.unity3d.com/Manual/SL-PlatformDifferences.html

        // Matrix4x4 leftView = Matrix4x4.identity;
        // Matrix4x4 rightView = Matrix4x4.identity;

        // if (SuperController.singleton.isOpenVR) {
        //   leftView = SuperController.singleton.ViveCenterCamera.GetStereoViewMatrix(Camera.StereoscopicEye.Left);
        //   rightView = SuperController.singleton.ViveCenterCamera.GetStereoViewMatrix(Camera.StereoscopicEye.Right);
        // } else if (SuperController.singleton.isOVR) {
        //   leftView = SuperController.singleton.OVRCenterCamera.GetStereoViewMatrix(Camera.StereoscopicEye.Left);
        //   rightView = SuperController.singleton.OVRCenterCamera.GetStereoViewMatrix(Camera.StereoscopicEye.Right);
        // } else {
        //   SuperController.LogError("view - yet another vr camera");
        // }

        // SuperController.LogMessage($"{leftView.GetRow(0)} {leftView.GetRow(1)} {leftView.GetRow(2)} {leftView.GetRow(3)}");

        // Matrix4x4[] inverseView = {
        //   (leftView).inverse,
        //   (rightView).inverse,
        // };

        var gpu_left = GL.GetGPUProjectionMatrix(leftProjection, true);
        var gpu_right = GL.GetGPUProjectionMatrix(rightProjection, true);

        Matrix4x4[] clipToWorld = {
          (gpu_left * postProcessingCamera.GetStereoViewMatrix(Camera.StereoscopicEye.Left)).inverse,
          (gpu_right * postProcessingCamera.GetStereoViewMatrix(Camera.StereoscopicEye.Right)).inverse,
        };

        // prepassMaterial.SetMatrixArray("_InverseView", inverseView);
        // prepassMaterial.SetMatrixArray("_InverseProjection", inverseProjection);
        prepassMaterial.SetMatrixArray("_ClipToWorld", clipToWorld);
      } else {
        // var viewMatrix = postProcessingCamera.worldToCameraMatrix.inverse;
        // var projectionMatrix = GL.GetGPUProjectionMatrix(postProcessingCamera.projectionMatrix, true).inverse;

        // projectionMatrix[1, 1] *= - 1;
        // projectionMatrix[1, 1] *= - 1;

        var ctw = (GL.GetGPUProjectionMatrix(postProcessingCamera.projectionMatrix, true) * postProcessingCamera.worldToCameraMatrix).inverse;

        // Matrix4x4[] inverseView = { viewMatrix, viewMatrix };
        // Matrix4x4[] inverseProjection = { projectionMatrix, projectionMatrix };
        Matrix4x4[] clipToWorld = { ctw, ctw };
      
        // prepassMaterial.SetMatrixArray("_InverseView", inverseView);
        // prepassMaterial.SetMatrixArray("_InverseProjection", inverseProjection);
        prepassMaterial.SetMatrixArray("_ClipToWorld", clipToWorld);
      }
    }
    
    private void OnDisable() {     
      // SuperController.LogMessage("EffectsCamera.OnDisable()");
      if (mainCamera != null)
        mainCamera.clearFlags = CameraClearFlags.SolidColor;
        mainCamera.cullingMask = DEFAULT_CULLING_MASK;

      postProcessingCamera.enabled = false;
    }

    private void OnDestroy() {
      // SuperController.LogMessage("EffectsCamera.OnDestroy()");
      postProcessingCamera.RemoveCommandBuffer(PRE_PROC_EVENT, prepassCommandBuffer);
      postProcessingCamera.RemoveCommandBuffer(POST_PROC_EVENT, commandBuffer);
      mainCamera.RemoveCommandBuffer(CameraEvent.AfterImageEffects, finalCommandBuffer);

      // CustomPostProcessing customPostProcessing = getOrAddCustomPostProcessingComponent();
      // DestroyImmediate(customPostProcessing);
      DestroyImmediate(postProcessingCamera);

      GameObject cameraContainer = GameObject.Find("Moyashi_EffectsCamera_CameraContainer");
      if (cameraContainer != null) DestroyImmediate(cameraContainer);      
    }
    
    private void OnPreCull() {
      postProcessingCamera.cullingMatrix = mainCamera.cullingMatrix;
    }

    private void recalculateSharpeningConsts() {
      float inputScreenX = postProcessingCamera.scaledPixelWidth;
      float inputScreenY = postProcessingCamera.scaledPixelHeight;

      float outputScreenX = postProcessingCamera.pixelWidth;
      float outputScreenY = postProcessingCamera.pixelHeight;

      uniforms.SharpeningConst0[0] = f32asuint32(inputScreenX * (1.0f / outputScreenX)); 
      uniforms.SharpeningConst0[1] = f32asuint32(inputScreenY * (1.0f / outputScreenY));
      uniforms.SharpeningConst0[2] = f32asuint32(0.5f * inputScreenX * (1.0f / outputScreenX) - 0.5f);
      uniforms.SharpeningConst0[3] = f32asuint32(0.5f * inputScreenY * (1.0f / outputScreenY) - 0.5f);

      float sharp = -(1.0f / (Mathf.Lerp(8.0f, 5.0f, uniforms.SharpeningScale)));
      Vector2 hSharp = new Vector2(sharp, 0.0f);

      uniforms.SharpeningConst1[0] = f32asuint32(sharp);
      uniforms.SharpeningConst1[1] = f32tof16(hSharp[0]) + (f32tof16(hSharp[1]) << 16); // pack 2x16 bit float
      uniforms.SharpeningConst1[2] = f32asuint32(8.0f * inputScreenX * (1.0f / outputScreenX));
      uniforms.SharpeningConst1[3] = 0.0f;
    }

    public uint f32tof16(float f) {
      ushort[] u16base = {
          0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
          0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
          0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
          0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
          0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000,
          0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0000, 0x0001, 0x0002, 0x0004, 0x0008, 0x0010, 0x0020, 0x0040, 0x0080, 0x0100, 0x0200, 0x0400,
          0x0800, 0x0c00, 0x1000, 0x1400, 0x1800, 0x1c00, 0x2000, 0x2400, 0x2800, 0x2c00, 0x3000, 0x3400, 0x3800, 0x3c00, 0x4000, 0x4400, 0x4800, 0x4c00, 0x5000,
          0x5400, 0x5800, 0x5c00, 0x6000, 0x6400, 0x6800, 0x6c00, 0x7000, 0x7400, 0x7800, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff,
          0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff,
          0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff,
          0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff,
          0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff,
          0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff,
          0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x7bff, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000,
          0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000,
          0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000,
          0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000,
          0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000,
          0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8000, 0x8001, 0x8002,
          0x8004, 0x8008, 0x8010, 0x8020, 0x8040, 0x8080, 0x8100, 0x8200, 0x8400, 0x8800, 0x8c00, 0x9000, 0x9400, 0x9800, 0x9c00, 0xa000, 0xa400, 0xa800, 0xac00,
          0xb000, 0xb400, 0xb800, 0xbc00, 0xc000, 0xc400, 0xc800, 0xcc00, 0xd000, 0xd400, 0xd800, 0xdc00, 0xe000, 0xe400, 0xe800, 0xec00, 0xf000, 0xf400, 0xf800,
          0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff,
          0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff,
          0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff,
          0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff,
          0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff,
          0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff, 0xfbff
      };

      byte[] u8shift = {
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x17, 0x16, 0x15, 0x14, 0x13, 0x12, 0x11, 0x10, 0x0f, 0x0e, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d,
          0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x17, 0x16, 0x15, 0x14, 0x13, 0x12, 0x11, 0x10, 0x0f, 0x0e, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d,
          0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x0d, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18,
          0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18, 0x18
      };

      f32uintUnion bits;
      bits.u = 0x000000;
      bits.f = f;
      uint u = bits.u;
      uint i = u >> 23;

      return (uint)((u16base[i]) + (u & 0x7fffff) >> u8shift[i]);
    }

    public uint f32asuint32(float f) {
      f32uintUnion bits;
      bits.u = 0x000000;
      bits.f = f;
      return bits.u;
    }
  }
}
